@@ -23,10 +23,10 @@ import (
 
 	configv1 "github.com/openshift/api/config/v1"
 
-	"github.com/openshift/cluster-version-operator/lib"
 	"github.com/openshift/cluster-version-operator/lib/resourcebuilder"
 	"github.com/openshift/cluster-version-operator/pkg/payload"
 	"github.com/openshift/cluster-version-operator/pkg/payload/precondition"
+	"github.com/openshift/library-go/pkg/manifest"
 )
 
 // ConfigSyncWorker abstracts how the image is synchronized to the server. Introduced for testing.
@@ -89,7 +89,8 @@ type SyncWorkerStatus struct {
 	Step    string
 	Failure error
 
-	Fraction float32
+	Done  int
+	Total int
 
 	Completed   int
 	Reconciling bool
@@ -155,11 +156,13 @@ type SyncWorker struct {
 	// manifests should be excluded based on an annotation
 	// of the form exclude.release.openshift.io/<identifier>=true
 	exclude string
+
+	clusterProfile string
 }
 
 // NewSyncWorker initializes a ConfigSyncWorker that will retrieve payloads to disk, apply them via builder
 // to a server, and obey limits about how often to reconcile or retry on errors.
-func NewSyncWorker(retriever PayloadRetriever, builder payload.ResourceBuilder, reconcileInterval time.Duration, backoff wait.Backoff, exclude string, eventRecorder record.EventRecorder) *SyncWorker {
+func NewSyncWorker(retriever PayloadRetriever, builder payload.ResourceBuilder, reconcileInterval time.Duration, backoff wait.Backoff, exclude string, eventRecorder record.EventRecorder, clusterProfile string) *SyncWorker {
 	return &SyncWorker{
 		retriever:     retriever,
 		builder:       builder,
@@ -175,14 +178,16 @@ func NewSyncWorker(retriever PayloadRetriever, builder payload.ResourceBuilder,
 		report: make(chan SyncWorkerStatus, 500),
 
 		exclude: exclude,
+
+		clusterProfile: clusterProfile,
 	}
 }
 
 // NewSyncWorkerWithPreconditions initializes a ConfigSyncWorker that will retrieve payloads to disk, apply them via builder
 // to a server, and obey limits about how often to reconcile or retry on errors.
 // It allows providing preconditions for loading payload.
-func NewSyncWorkerWithPreconditions(retriever PayloadRetriever, builder payload.ResourceBuilder, preconditions precondition.List, reconcileInterval time.Duration, backoff wait.Backoff, exclude string, eventRecorder record.EventRecorder) *SyncWorker {
-	worker := NewSyncWorker(retriever, builder, reconcileInterval, backoff, exclude, eventRecorder)
+func NewSyncWorkerWithPreconditions(retriever PayloadRetriever, builder payload.ResourceBuilder, preconditions precondition.List, reconcileInterval time.Duration, backoff wait.Backoff, exclude string, eventRecorder record.EventRecorder, clusterProfile string) *SyncWorker {
+	worker := NewSyncWorker(retriever, builder, reconcileInterval, backoff, exclude, eventRecorder, clusterProfile)
 	worker.preconditions = preconditions
 	return worker
 }
@@ -370,15 +375,23 @@ type statusWrapper struct {
 
 func (w *statusWrapper) Report(status SyncWorkerStatus) {
 	p := w.previousStatus
+	var fractionComplete float32
+	if status.Total > 0 {
+		fractionComplete = float32(status.Done) / float32(status.Total)
+	}
+	var previousFractionComplete float32
+	if p.Total > 0 {
+		previousFractionComplete = float32(p.Done) / float32(p.Total)
+	}
 	if p.Failure != nil && status.Failure == nil {
 		if p.Actual.Image == status.Actual.Image {
-			if status.Fraction < p.Fraction {
+			if fractionComplete < previousFractionComplete {
 				klog.V(5).Infof("Dropping status report from earlier in sync loop")
 				return
 			}
 		}
 	}
-	if status.Fraction > p.Fraction || status.Completed > p.Completed || (status.Failure == nil && status.Actual.Image != p.Actual.Image) {
+	if fractionComplete > previousFractionComplete || status.Completed > p.Completed || (status.Failure == nil && status.Actual.Image != p.Actual.Image) {
 		status.LastProgress = time.Now()
 	}
 	if status.Generation == 0 {
@@ -429,12 +442,35 @@ func (w *SyncWorker) calculateNext(work *SyncWork) bool {
 	return changed
 }
 
-// equalUpdate returns true if two updates have the same image.
+// equalUpdate returns true if two updates are semantically equivalent.
+// It checks if the updates have have the same force and image values.
+//
+// We require complete pullspec equality, not just digest equality,
+// because we want to go through the usual update process even if it's
+// only the registry portion of the pullspec which has changed.
 func equalUpdate(a, b configv1.Update) bool {
-	if a.Force != b.Force {
-		return false
+	return a.Force == b.Force && a.Image == b.Image
+}
+
+// equalDigest returns true if and only if the two pullspecs are
+// by-digest pullspecs and the digests are equal or the two pullspecs
+// are identical.
+func equalDigest(pullspecA string, pullspecB string) bool {
+	if pullspecA == pullspecB {
+		return true
 	}
-	return a.Image == b.Image
+	digestA := splitDigest(pullspecA)
+	return digestA != "" && digestA == splitDigest(pullspecB)
+}
+
+// splitDigest returns the pullspec's digest, or an empty string if
+// the pullspec is not by-digest (e.g. it is by-tag, or implicit).
+func splitDigest(pullspec string) string {
+	parts := strings.SplitN(pullspec, "@", 3)
+	if len(parts) != 2 {
+		return ""
+	}
+	return parts[1]
 }
 
 // equalSyncWork returns true if a and b are equal.
@@ -512,6 +548,10 @@ func (w *SyncWorker) syncOnce(ctx context.Context, work *SyncWork, maxWorkers in
 	}
 	klog.V(4).Infof("Running sync %s (force=%t) on generation %d in state %s at attempt %d", versionString(desired), work.Desired.Force, work.Generation, work.State, work.Attempt)
 
+	if work.Attempt == 0 {
+		payload.InitCOUpdateStartTimes()
+	}
+
 	// cache the payload until the release image changes
 	validPayload := w.payload
 	if validPayload != nil && validPayload.Release.Image == desired.Image {
@@ -543,7 +583,7 @@ func (w *SyncWorker) syncOnce(ctx context.Context, work *SyncWork, maxWorkers in
 		}
 
 		w.eventRecorder.Eventf(cvoObjectRef, corev1.EventTypeNormal, "VerifyPayload", "verifying payload version=%q image=%q", desired.Version, desired.Image)
-		payloadUpdate, err := payload.LoadUpdate(info.Directory, desired.Image, w.exclude)
+		payloadUpdate, err := payload.LoadUpdate(info.Directory, desired.Image, w.exclude, w.clusterProfile)
 		if err != nil {
 			w.eventRecorder.Eventf(cvoObjectRef, corev1.EventTypeWarning, "VerifyPayloadFailed", "verifying payload failed version=%q image=%q failure=%v", desired.Version, desired.Image, err)
 			reporter.Report(SyncWorkerStatus{
@@ -699,6 +739,11 @@ func (w *SyncWorker) apply(ctx context.Context, payloadUpdate *payload.Update, w
 				if task.Manifest.GVK != configv1.SchemeGroupVersion.WithKind("ClusterOperator") {
 					continue
 				}
+				ov, ok := getOverrideForManifest(work.Overrides, task.Manifest)
+				if ok && ov.Unmanaged {
+					klog.V(4).Infof("Skipping precreation of %s as unmanaged", task)
+					continue
+				}
 				if err := w.builder.Apply(ctx, task.Manifest, payload.PrecreatingPayload); err != nil {
 					klog.V(2).Infof("Unable to precreate resource %s: %v", task, err)
 					continue
@@ -789,7 +834,8 @@ func (r *consistentReporter) Update() {
 	metricPayload.WithLabelValues(r.version, "applied").Set(float64(r.done))
 	copied := r.status
 	copied.Step = "ApplyResources"
-	copied.Fraction = float32(r.done) / float32(r.total)
+	copied.Done = r.done
+	copied.Total = r.total
 	r.reporter.Report(copied)
 }
 
@@ -798,7 +844,8 @@ func (r *consistentReporter) Error(err error) {
 	defer r.lock.Unlock()
 	copied := r.status
 	copied.Step = "ApplyResources"
-	copied.Fraction = float32(r.done) / float32(r.total)
+	copied.Done = r.done
+	copied.Total = r.total
 	if !isContextError(err) {
 		copied.Failure = err
 	}
@@ -812,7 +859,8 @@ func (r *consistentReporter) Errors(errs []error) error {
 	defer r.lock.Unlock()
 	copied := r.status
 	copied.Step = "ApplyResources"
-	copied.Fraction = float32(r.done) / float32(r.total)
+	copied.Done = r.done
+	copied.Total = r.total
 	if err != nil {
 		copied.Failure = err
 	}
@@ -835,7 +883,8 @@ func (r *consistentReporter) Complete() {
 	copied.Completed = r.completed + 1
 	copied.Initial = false
 	copied.Reconciling = true
-	copied.Fraction = 1
+	copied.Done = r.done
+	copied.Total = r.total
 	r.reporter.Report(copied)
 }
 
@@ -940,6 +989,7 @@ func isClusterOperatorNotAvailable(err error) bool {
 // newClusterOperatorsNotAvailable unifies multiple ClusterOperatorNotAvailable errors into
 // a single error. It returns nil if the provided errors are not of the same type.
 func newClusterOperatorsNotAvailable(errs []error) error {
+	updateEffect := payload.UpdateEffectNone
 	names := make([]string, 0, len(errs))
 	for _, err := range errs {
 		uErr, ok := err.(*payload.UpdateError)
@@ -949,6 +999,15 @@ func newClusterOperatorsNotAvailable(errs []error) error {
 		if len(uErr.Name) > 0 {
 			names = append(names, uErr.Name)
 		}
+		switch uErr.UpdateEffect {
+		case payload.UpdateEffectNone:
+		case payload.UpdateEffectFail:
+			updateEffect = payload.UpdateEffectFail
+		case payload.UpdateEffectFailAfterInterval:
+			if updateEffect != payload.UpdateEffectFail {
+				updateEffect = payload.UpdateEffectFailAfterInterval
+			}
+		}
 	}
 	if len(names) == 0 {
 		return nil
@@ -961,10 +1020,11 @@ func newClusterOperatorsNotAvailable(errs []error) error {
 	sort.Strings(names)
 	name := strings.Join(names, ", ")
 	return &payload.UpdateError{
-		Nested:  errors.NewAggregate(errs),
-		Reason:  "ClusterOperatorsNotAvailable",
-		Message: fmt.Sprintf("Some cluster operators are still updating: %s", name),
-		Name:    name,
+		Nested:       errors.NewAggregate(errs),
+		UpdateEffect: updateEffect,
+		Reason:       "ClusterOperatorsNotAvailable",
+		Message:      fmt.Sprintf("Some cluster operators are still updating: %s", name),
+		Name:         name,
 	}
 }
 
@@ -1014,9 +1074,9 @@ func newMultipleError(errs []error) error {
 }
 
 // getOverrideForManifest returns the override and true when override exists for manifest.
-func getOverrideForManifest(overrides []configv1.ComponentOverride, manifest *lib.Manifest) (configv1.ComponentOverride, bool) {
+func getOverrideForManifest(overrides []configv1.ComponentOverride, manifest *manifest.Manifest) (configv1.ComponentOverride, bool) {
 	for idx, ov := range overrides {
-		kind, namespace, name := manifest.GVK.Kind, manifest.Object().GetNamespace(), manifest.Object().GetName()
+		kind, namespace, name := manifest.GVK.Kind, manifest.Obj.GetNamespace(), manifest.Obj.GetName()
 		if ov.Kind == kind &&
 			(namespace == "" || ov.Namespace == namespace) && // cluster-scoped objects don't have namespace.
 			ov.Name == name {
